from dataclasses import dataclass
from enum import Enum


@dataclass
class Position:
    szi: int
    entry_ntl: int
    isolated_raw_usd: int
    leverage: int
    is_isolated: bool


@dataclass
class SpotBalance:
    total: int
    hold: int
    entry_ntl: int


@dataclass
class UserVaultEquity:
    equity: int
    locked_until_timestamp: int


@dataclass
class Delegation:
    validator: str
    amount: int
    locked_until_timestamp: int


@dataclass
class DelegatorSummary:
    delegated: int
    undelegated: int
    total_pending_withdrawal: int
    n_pending_withdrawals: int


@dataclass
class PerpAssetInfo:
    coin: str
    margin_table_id: int
    sz_decimals: int
    max_leverage: int
    only_isolated: bool


@dataclass
class SpotInfo:
    name: str
    tokens: list[int]

    def __post_init__(self):
        assert len(self.tokens) == 2, (
            "SpotInfo.tokens length must be 2"
        )  # prama: no cover


@dataclass
class TokenInfo:
    name: str
    spots: list[int]
    deployer_trading_fee_share: int
    deployer: str
    evm_contract: str
    sz_decimals: int
    wei_decimals: int
    evm_extra_wei_decimals: int


@dataclass
class UserBalance:
    user: str
    balance: int


@dataclass
class TokenSupply:
    max_supply: int
    total_supply: int
    circulating_supply: int
    future_emissions: int
    non_circulating_user_balances: list[UserBalance]


@dataclass
class Bbo:
    bid: int
    ask: int


@dataclass
class AccountMarginSummary:
    account_value: int
    margin_used: int
    ntl_pos: int
    raw_usd: int


class ActionType(Enum):
    LimitOrder = 1
    VaultTransfer = 2
    TokenDelegate = 3
    StakingDeposit = 4
    StakingWithdraw = 5
    SpotSend = 6
    USDClassTransfer = 7
    FinalizeEVMContract = 8
    AddAPIWallet = 9
    CancelOrderByOid = 10
    CancelOrderByCloid = 11


class Tif(Enum):
    Alo = 1
    Gtc = 2
    Ioc = 3


class AssetType(str, Enum):
    PERP = "perp"
    SPOT = "spot"


class FinalizeEvmContractVariant(Enum):
    Create = 1
    FirstStorageSlot = 2
    CustomStorageSlot = 3



from logging import getLogger

from web3 import Web3
from eth_abi import abi
from eth_account import Account

from hl_web3.utils.miscs import load_abi, get_raw_action
from hl_web3.utils.constants import HL_CORE_WRITER

from .endpoint import Endpoint
from .utils.types import Tif, ActionType, FinalizeEvmContractVariant


class Exchange(Endpoint):
    def __init__(self, rpc_url: str, private_key: str):
        super().__init__(rpc_url)
        self.account = Account.from_key(private_key)
        self.core_writer = self._w3.eth.contract(
            address=Web3.to_checksum_address(HL_CORE_WRITER),
            abi=load_abi("CoreWriter"),
        )
        self.logger = getLogger("hl_web3.exchange")

    async def _send_raw_action(self, action_type: ActionType, action: bytes):
        raw_action = get_raw_action(action_type, action)

        self.logger.debug(f"Sending raw action: {raw_action.hex()}")

        data = self.core_writer.encode_abi("sendRawAction", args=[raw_action])
        basic_tx = {
            "to": HL_CORE_WRITER,
            "data": data,
            "from": self.account.address,
        }
        tx = {
            **basic_tx,
            "gas": await self._w3.eth.estimate_gas(basic_tx),  # type: ignore
            "gasPrice": await self._w3.eth.gas_price,
            "nonce": await self._w3.eth.get_transaction_count(
                self.account.address
            ),
            "chainId": await self._w3.eth.chain_id,
        }
        signed_tx = self._w3.eth.account.sign_transaction(tx, self.account.key)

        self.logger.debug(f"Sending tx: {signed_tx.raw_transaction.hex()}")

        tx_hash = await self._w3.eth.send_raw_transaction(
            signed_tx.raw_transaction
        )
        return tx_hash.hex()

    # action 1
    # Cloid encoding: 0 means no cloid, otherwise uses the number as the cloid. Px and sz should be sent as 10^8 * the human readable value
    async def place_order(
        self,
        asset: int,
        is_buy: bool,
        px: int,
        sz: int,
        ro: bool,
        tif: Tif,
        cloid: int = 0,
    ):
        action = abi.encode(
            ["uint32", "bool", "uint64", "uint64", "bool", "uint8", "uint128"],
            [asset, is_buy, px, sz, ro, tif.value, cloid],
        )
        return await self._send_raw_action(ActionType.LimitOrder, action)

    # action 2
    async def vault_transfer(self, vault: str, is_deposit: bool, usd: int):
        action = abi.encode(
            ["address", "bool", "uint64"], [vault, is_deposit, usd]
        )
        return await self._send_raw_action(ActionType.VaultTransfer, action)

    # action 3
    async def token_delegate(
        self, validator: str, amount: int, is_undelegate: bool
    ):
        action = abi.encode(
            ["address", "uint64", "bool"], [validator, amount, is_undelegate]
        )
        return await self._send_raw_action(ActionType.TokenDelegate, action)

    # action 4
    async def staking_deposit(self, amount: int):
        action = abi.encode(["uint64"], [amount])
        return await self._send_raw_action(ActionType.StakingDeposit, action)

    # action 5
    async def staking_withdraw(self, amount: int):
        action = abi.encode(["uint64"], [amount])
        return await self._send_raw_action(ActionType.StakingWithdraw, action)

    # action 6
    async def spot_send(self, dest: str, token: int, amount: int):
        action = abi.encode(
            ["address", "uint64", "uint64"], [dest, token, amount]
        )
        return await self._send_raw_action(ActionType.SpotSend, action)

    # action 7
    async def send_usd_class_transfer(self, ntl: int, to_perp: bool):
        action = abi.encode(["uint64", "bool"], [ntl, to_perp])
        return await self._send_raw_action(ActionType.USDClassTransfer, action)

    # action 8
    # status: 1 for Create, 2 for FirstStorageSlot, 3 for CustomStorageSlot. If Create variant, then createNonce input argument is used.
    async def finalize_evm_contract(
        self,
        token: int,
        variant: FinalizeEvmContractVariant,
        create_nonce: int = 0,
    ):
        if variant == FinalizeEvmContractVariant.Create:
            assert create_nonce > 0, (
                "create_nonce must be greater than 0 for Create variant"
            )

        action = abi.encode(
            ["uint64", "uint8", "uint64"], [token, variant.value, create_nonce]
        )
        return await self._send_raw_action(
            ActionType.FinalizeEVMContract, action
        )

    # action 9
    # If the API wallet name is empty then this becomes the main API wallet / agent
    async def add_api_wallet(self, wallet: str, name: str):
        action = abi.encode(["address", "string"], [wallet, name])
        return await self._send_raw_action(ActionType.AddAPIWallet, action)

    # action 10
    async def cancel_order_by_oid(self, asset: int, oid: int):
        action = abi.encode(["uint32", "uint64"], [asset, oid])
        return await self._send_raw_action(ActionType.CancelOrderByOid, action)

    # action 11
    async def cancel_order_by_cloid(self, asset: int, cloid: int):
        action = abi.encode(["uint32", "uint128"], [asset, cloid])
        return await self._send_raw_action(
            ActionType.CancelOrderByCloid, action
        )



from logging import getLogger

from web3 import AsyncWeb3


class Endpoint:
    def __init__(self, rpc_url: str):
        if rpc_url.startswith("http"):
            provider = AsyncWeb3.AsyncHTTPProvider(rpc_url)
        elif rpc_url.startswith("ws"):
            provider = AsyncWeb3.WebSocketProvider(rpc_url)
        else:
            raise ValueError(f"Invalid RPC URL: {rpc_url}")

        self._w3 = AsyncWeb3(provider)  # type: ignore
        self.logger = getLogger("hl_web3.endpoint")

    async def _call(self, tx: dict):
        self.logger.debug(f"Eth call: {tx}")
        resp = await self._w3.eth.call(tx)  # type: ignore
        self.logger.debug(f"Eth call response: {resp.hex()}")
        return resp



from logging import getLogger

from eth_abi import abi

import hl_web3.utils.constants as constants
from hl_web3.utils.types import (
    AccountMarginSummary,
    AssetType,
    Bbo,
    Delegation,
    DelegatorSummary,
    PerpAssetInfo,
    Position,
    SpotBalance,
    SpotInfo,
    TokenInfo,
    TokenSupply,
    UserVaultEquity,
)

from .endpoint import Endpoint
from .utils.abi_types import (
    AccountMarginSummaryABI,
    BboABI,
    DelegationsABI,
    DelegatorSummaryABI,
    PerpAssetInfoABI,
    PositionABI,
    SpotBalanceABI,
    SpotInfoABI,
    TokenInfoABI,
    TokenSupplyABI,
    UserVaultEquityABI,
)


class Info(Endpoint):
    def __init__(self, rpc_url: str):
        super().__init__(rpc_url)
        self.logger = getLogger("hl_web3.info")

    async def get_user_position(self, user: str, asset: int):
        data = abi.encode(["address", "uint16"], [user, asset])
        tx = {"to": constants.HL_POSITION, "data": data}
        resp = await self._call(tx)
        (pos,) = abi.decode([PositionABI], resp)
        return Position(*pos)

    async def get_user_spot_balance(self, user: str, token: int):
        data = abi.encode(["address", "uint64"], [user, token])
        tx = {"to": constants.HL_SPOT_BALANCE, "data": data}
        resp = await self._call(tx)
        (balance,) = abi.decode([SpotBalanceABI], resp)
        return SpotBalance(*balance)

    async def get_user_vault_equity(self, user: str, vault: str):
        data = abi.encode(["address", "address"], [user, vault])
        tx = {"to": constants.HL_VAULT_EQUITY, "data": data}
        resp = await self._call(tx)
        (equity,) = abi.decode([UserVaultEquityABI], resp)
        return UserVaultEquity(*equity)

    async def get_user_withdrawable(self, user: str):
        data = abi.encode(["address"], [user])
        tx = {"to": constants.HL_WITHDRAWABLE, "data": data}
        resp = await self._call(tx)
        (amount,) = abi.decode(["uint64"], resp)

        return amount / constants.USD_SCALE_FACTOR

    async def get_user_delegations(self, user: str):
        data = abi.encode(["address"], [user])
        tx = {"to": constants.HL_DELEGATIONS, "data": data}
        resp = await self._call(tx)

        (delegations,) = abi.decode([DelegationsABI], resp)

        return [Delegation(*delegation) for delegation in delegations]

    async def get_user_delegator_summary(self, user: str):
        data = abi.encode(["address"], [user])
        tx = {"to": constants.HL_DELEGATOR_SUMMARY, "data": data}
        resp = await self._call(tx)
        (summary,) = abi.decode([DelegatorSummaryABI], resp)
        return DelegatorSummary(*summary)

    async def get_mark_px(self, asset: int):
        data = abi.encode(["uint32"], [asset])
        tx = {"to": constants.HL_MARK_PX, "data": data}
        resp = await self._call(tx)
        (px,) = abi.decode(["uint64"], resp)

        scale_factor = await self.get_asset_scale_factor(asset, AssetType.PERP)

        return px / scale_factor

    async def get_oracle_px(self, asset: int):
        data = abi.encode(["uint32"], [asset])
        tx = {"to": constants.HL_ORACLE_PX, "data": data}
        resp = await self._call(tx)
        (px,) = abi.decode(["uint64"], resp)

        scale_factor = await self.get_asset_scale_factor(asset, AssetType.PERP)

        return px / scale_factor

    async def get_spot_px(self, asset: int):
        data = abi.encode(["uint32"], [asset])
        tx = {"to": constants.HL_SPOT_PX, "data": data}
        resp = await self._call(tx)
        (px,) = abi.decode(["uint64"], resp)

        scale_factor = await self.get_asset_scale_factor(asset, AssetType.SPOT)

        return px / scale_factor

    async def get_block_number(self):
        tx = {"to": constants.HL_L1_BLOCK_NUMBER, "data": bytes()}
        resp = await self._call(tx)
        (blk_number,) = abi.decode(["uint64"], resp)
        return blk_number

    async def get_perp_asset_info(self, perp: int):
        data = abi.encode(["uint32"], [perp])
        tx = {"to": constants.HL_PERP_ASSET_INFO, "data": data}
        resp = await self._call(tx)
        (perp_asset,) = abi.decode([PerpAssetInfoABI], resp)
        return PerpAssetInfo(*perp_asset)

    async def get_spot_info(self, spot: int):
        data = abi.encode(["uint32"], [spot])
        tx = {"to": constants.HL_SPOT_INFO, "data": data}
        resp = await self._call(tx)
        (spot_info,) = abi.decode([SpotInfoABI], resp)
        return SpotInfo(*spot_info)

    async def get_token_info(self, token: int):
        data = abi.encode(["uint32"], [token])
        tx = {"to": constants.HL_TOKEN_INFO, "data": data}
        resp = await self._call(tx)
        (token_info,) = abi.decode([TokenInfoABI], resp)
        return TokenInfo(*token_info)

    async def get_token_supply(self, token: int):
        data = abi.encode(["uint32"], [token])
        tx = {"to": constants.HL_TOKEN_SUPPLY, "data": data}
        resp = await self._call(tx)
        (supply,) = abi.decode([TokenSupplyABI], resp)
        return TokenSupply(*supply)

    async def get_bbo(self, asset: int):
        data = abi.encode(["uint32"], [asset])
        tx = {"to": constants.HL_BBO, "data": data}
        resp = await self._call(tx)
        (bbo,) = abi.decode([BboABI], resp)
        return Bbo(*bbo)

    async def get_user_margin_summary(self, dex: int, user: str):
        data = abi.encode(["uint32", "address"], [dex, user])
        tx = {"to": constants.HL_ACCOUNT_MARGIN_SUMMARY, "data": data}
        resp = await self._call(tx)
        (summary,) = abi.decode([AccountMarginSummaryABI], resp)
        return AccountMarginSummary(*summary)

    async def is_core_user(self, user: str):
        data = abi.encode(["address"], [user])
        tx = {"to": constants.HL_CORE_USER, "data": data}
        resp = await self._call(tx)
        (is_core_user,) = abi.decode(["bool"], resp)
        return is_core_user

    async def get_asset_scale_factor(self, asset: int, asset_type: AssetType):
        # based on https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/hyperevm/interacting-with-hypercore#read-precompiles
        if asset_type == AssetType.PERP:
            perp_asset = await self.get_perp_asset_info(asset)
            exponent = 6 - perp_asset.sz_decimals
        elif asset_type == AssetType.SPOT:
            spot_info = await self.get_spot_info(asset)
            (_, base_token) = spot_info.tokens
            base_token_info = await self.get_token_info(base_token)
            exponent = 8 - base_token_info.sz_decimals
        else:  # pragma: no cover
            raise ValueError(f"Invalid asset type: {asset_type}")

        return 10**exponent



PositionABI = "(int64,uint64,int64,uint32,bool)"
SpotBalanceABI = "(uint64,uint64,uint64)"
UserVaultEquityABI = "(uint64,uint64)"
DelegationsABI = "(address,uint64,uint64)[]"
DelegatorSummaryABI = "(uint64,uint64,uint64,uint64)"
PerpAssetInfoABI = "(string,uint32,uint8,uint8,bool)"
SpotInfoABI = "(string,uint64[2])"
TokenInfoABI = "(string,uint64[],uint64,address,address,uint8,uint8,int8)"
TokenSupplyABI = "(uint64,uint64,uint64,uint64,(address,uint64)[])"
BboABI = "(uint64,uint64)"
AccountMarginSummaryABI = "(int64,uint64,uint64,int64)"



import json

from .constants import ACTION_VERSION
from .types import ActionType


def get_raw_action(action_type: ActionType, action: bytes | str) -> bytes:
    header = bytes([ACTION_VERSION])

    action_id = action_type.value
    header += action_id.to_bytes(3, "big")

    if isinstance(action, str):
        action = bytes.fromhex(action.lstrip("0x"))

    return header + action


def load_abi(name: str):
    with open(f"abis/{name}.json", "r") as f:
        return json.load(f)
